@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2015, University of Amsterdam
+    Copyright (c)  1985-2017, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -36,6 +36,8 @@
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #include "os/pl-ctype.h"
+#undef LD
+#define LD LOCAL_LD
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Implementation issues
@@ -162,8 +164,8 @@ This is a hard problem. Atom-GC cannot   run  while some thread performs
 normal GC because the pointer relocation makes it extremely hard to find
 referenced atoms. Otherwise, ask all  threads   to  mark their reachable
 atoms and run collectAtoms() to reclaim the unreferenced atoms. The lock
-L_GC is used to ensure garbage collection does not run concurrently with
-atom garbage collection.
+LD->thread.scan_lock is used to ensure garbage   collection does not run
+concurrently with atom garbage collection.
 
 Atom-GC asynchronously walks  the  stacks  of   all  threads  and  marks
 everything  that  looks  `atom-like',   i.e.,    our   collector   is  a
@@ -176,14 +178,8 @@ Note that threads can mark their atoms and continue execution because:
 
   - If a marked atom is no longer needed it is merely not reclaimed this
     time (but might be in the next collection).
-  - If a new atom is referenced from the stack it is either a
-    - builtin atom (no problem)
-    - an atom from a structure using reference counting (is referenced
-      by this structure, so no problem, unless the reference count drops
-      to zero in PL_unregister_atom().  See PL_unregister_atom() for
-      handling the no-locking case.
-    - It is created.  This case blocks on L_ATOM being locked from
-      lookupBlob().
+  - New atoms are added to the stacks using pushVolatileAtom(), which
+    marks the atom if AGC is active.
   - Finally, message queues and bags as used by findall/3 complicate
     the issue.  An atom sent to these structures subsequently may
     become inaccessible from the stack (the normal case for findall/3,
@@ -198,30 +194,30 @@ Note that threads can mark their atoms and continue execution because:
       AGC is running, we are ok, because this is merely the same issue
       as atoms living on the stack.  TBD: redesign the structures such
       that they can safely be walked.
-
-JW: I think we can reduce locking for AGC further.
-  - L_GC is needed, but we could give each thread its own GC lock,
-    as we only need to sync with the thread we are marking.
-  - L_THREAD is needed to guarantee no new threads are started or
-    destroyed. That too can probably be done more subtle.
-  - L_AGC is only used for syncing with current_blob().
-  - Why is L_ATOM needed?
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static void	rehashAtoms(void);
+static int	rehashAtoms(void);
+static void	considerAGC(void);
 
-#define atomTable    GD->atoms.table
+static inline int
+bump_atom_references(Atom a, unsigned int ref)
+{ for(;;)
+  { unsigned int nref = ref+1;
 
-#if O_DEBUG
-#define lookups GD->atoms.lookups
-#define	cmps	GD->atoms.cmps
-#endif
-
-#define LOCK()   PL_LOCK(L_ATOM)
-#define UNLOCK() PL_UNLOCK(L_ATOM)
-#undef LD
-#define LD LOCAL_LD
+    if ( ATOM_REF_COUNT(nref) == 0 )
+      return TRUE;			/* reached max references */
 
+    if ( COMPARE_AND_SWAP(&a->references, ref, nref) )
+    { if ( ATOM_REF_COUNT(ref) == 0 )
+	ATOMIC_DEC(&GD->atoms.unregistered);
+      return TRUE;
+    } else
+    { ref = a->references;
+      if ( !ATOM_IS_VALID(ref) )
+	return FALSE;
+    }
+  }
+}
 
 		 /*******************************
 		 *	      TYPES		*
@@ -243,33 +239,37 @@ static PL_blob_t unregistered_blob_atom =
 
 void
 PL_register_blob_type(PL_blob_t *type)
-{ PL_LOCK(L_MISC);			/* cannot use L_ATOM */
+{ if ( !type->registered )
+  { PL_LOCK(L_MISC);
 
-  if ( !type->registered )
-  { if ( !GD->atoms.types )
-    { GD->atoms.types = type;
-    } else
-    { PL_blob_t *t = GD->atoms.types;
+    if ( !type->registered )
+    { if ( !GD->atoms.types )
+      { GD->atoms.types = type;
+      } else
+      { PL_blob_t *t = GD->atoms.types;
 
-      while(t->next)
-	t = t->next;
+	while(t->next)
+	  t = t->next;
 
-      t->next = type;
-      type->rank = t->rank+1;
+	t->next = type;
+	type->rank = t->rank+1;
+      }
+      if ( true(type, PL_BLOB_TEXT) )
+      { if ( true(type, PL_BLOB_WCHAR) )
+	  type->padding = sizeof(pl_wchar_t);
+	else
+	  type->padding = sizeof(char);
+      }
+
+      if ( !GD->atoms.initialised )
+	type->registered = TRUE;
+      if ( !type->atom_name )
+	type->atom_name = PL_new_atom(type->name);
+      type->registered = TRUE;
     }
-    type->registered = TRUE;
-    if ( !type->atom_name )
-      type->atom_name = PL_new_atom(type->name);
 
-    if ( true(type, PL_BLOB_TEXT) )
-    { if ( true(type, PL_BLOB_WCHAR) )
-	type->padding = sizeof(pl_wchar_t);
-      else
-	type->padding = sizeof(char);
-    }
+    PL_UNLOCK(L_MISC);
   }
-
-  PL_UNLOCK(L_MISC);
 }
 
 
@@ -307,30 +307,32 @@ PL_unregister_blob_type(PL_blob_t *type)
 
   PL_register_blob_type(&unregistered_blob_atom);
 
-  LOCK();
   for(index=1, i=0; !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
     for(; index<upto; index++)
     { Atom atom = b + index;
+      unsigned int refs = atom->references;
+      PL_blob_t *btype = atom->type;
 
-      if ( ATOM_IS_VALID(atom->references) && atom->type == type )
+      if ( ATOM_IS_VALID(refs) && btype == type &&
+	   bump_atom_references(atom, refs) )
       { atom->type = &unregistered_blob_atom;
 
 	atom->name = "<discarded blob>";
 	atom->length = strlen(atom->name);
-
 	discarded++;
+	PL_unregister_atom(atom->atom);
       }
     }
   }
-  UNLOCK();
 
   return discarded == 0 ? TRUE : FALSE;
 }
@@ -352,7 +354,7 @@ static const ccharp atoms[] = {
 #ifdef O_PLMT
 
 #define acquire_atom_table(t, b) \
-  { LD->thread.info->access.atom_table = atomTable; \
+  { LD->thread.info->access.atom_table = GD->atoms.table; \
     t = LD->thread.info->access.atom_table->table; \
     b = LD->thread.info->access.atom_table->buckets; \
   }
@@ -373,8 +375,8 @@ static const ccharp atoms[] = {
 #else
 
 #define acquire_atom_table(t, b) \
-  { t = atomTable->table; \
-    b = atomTable->buckets; \
+  { t = GD->atoms.table->table; \
+    b = GD->atoms.table->buckets; \
   }
 
 #define release_atom_table() (void)0
@@ -399,21 +401,23 @@ another atom.
 
 static void
 allocateAtomBlock(int idx)
-{
-  PL_LOCK(L_MISC);
-
-  if ( !GD->atoms.array.blocks[idx] )
+{ if ( !GD->atoms.array.blocks[idx] )
   { size_t bs = (size_t)1<<idx;
+    size_t i;
     Atom newblock;
 
     if ( !(newblock=PL_malloc_uncollectable(bs*sizeof(struct atom))) )
       outOfCore();
 
     memset(newblock, 0, bs*sizeof(struct atom));
-    GD->atoms.array.blocks[idx] = newblock-bs;
+    for(i=0; i<bs; i++)
+    { newblock[i].type = ATOM_TYPE_INVALID;
+      newblock[i].name = "<virgin>";
+    }
+    if ( !COMPARE_AND_SWAP(&GD->atoms.array.blocks[idx],
+			   NULL, newblock-bs) )
+      PL_free(newblock);		/* done by someone else */
   }
-
-  PL_UNLOCK(L_MISC);
 }
 
 static Atom
@@ -428,21 +432,22 @@ reserveAtom(void)
 
   for(index=GD->atoms.no_hole_before, i=MSB(index); !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
     for(; index<upto; index++)
-    {
-      a = b + index;
+    { a = b + index;
       ref = a->references;
 
       if ( ATOM_IS_FREE(ref) &&
 	   COMPARE_AND_SWAP(&a->references, ref, ATOM_RESERVED_REFERENCE) )
-      { GD->atoms.no_hole_before = index+1;
+      { assert(a->type == ATOM_TYPE_INVALID);
+	GD->atoms.no_hole_before = index+1;
         a->atom = (index<<LMASK_BITS)|TAG_ATOM;
 
 	return a;
@@ -452,28 +457,25 @@ reserveAtom(void)
   GD->atoms.no_hole_before = index+1;
 #endif /*O_ATOMGC*/
 
-redo:
+  for(;;)
+  { index = GD->atoms.highest;
+    idx = MSB(index);
+    assert(index >= 0);
 
-  index = GD->atoms.highest;
-  idx = MSB(index);
-  assert(index >= 0);
+    if ( !GD->atoms.array.blocks[idx] )
+      allocateAtomBlock(idx);
 
-  if ( !GD->atoms.array.blocks[idx] )
-  { allocateAtomBlock(idx);
+    a = &GD->atoms.array.blocks[idx][index];
+    ref = a->references;
+
+    if ( ATOM_IS_FREE(ref) &&
+	 COMPARE_AND_SWAP(&a->references, ref, ATOM_RESERVED_REFERENCE) )
+    { COMPARE_AND_SWAP(&GD->atoms.highest, index, index+1);
+      a->atom = (index<<LMASK_BITS)|TAG_ATOM;
+
+      return a;
+    }
   }
-
-  a = &GD->atoms.array.blocks[idx][index];
-  ref = a->references;
-
-  if ( ATOM_IS_FREE(ref) &&
-       COMPARE_AND_SWAP(&a->references, ref, ATOM_RESERVED_REFERENCE) )
-  { ATOMIC_INC(&GD->atoms.highest);
-    a->atom = (index<<LMASK_BITS)|TAG_ATOM;
-
-    return a;
-  }
-
-  goto redo;
 }
 
 
@@ -492,8 +494,25 @@ pick up the request and process it.
 
 PL_handle_signals() decides on the actual invocation of atom-gc and will
 treat the signal as bogus if agc has already been performed.
+
+(**)   Without   this    check,    some     threads    may    pass   the
+PL_LOCK(L_REHASH_ATOMS) around rehashAtoms() and create   their atom. If
+they manage to register the atom in   the old table before rehashAtoms()
+activates the new table the insertion   is successful, but rehashAtoms()
+may not have moved the atom to the new   table. Now we will repeat if we
+bypassed the LOCK as either GD->atoms.rehashing is TRUE or the new table
+is activated.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+static int
+same_name(const Atom a, const char *s, size_t length, const PL_blob_t *type)
+{ if ( false(type, PL_BLOB_NOCOPY) )
+    return memcmp(s, a->name, length) == 0;
+  else
+    return s == a->name;
+}
+
+
 word
 lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
 { GET_LD
@@ -513,85 +532,43 @@ redo:
   v  = v0 & (buckets-1);
   head = table[v];
   acquire_atom_bucket(table+v);
-  DEBUG(MSG_HASH_STAT, lookups++);
+  DEBUG(MSG_HASH_STAT, GD->atoms.lookups++);
 
   if ( true(type, PL_BLOB_UNIQUE) )
-  { if ( false(type, PL_BLOB_NOCOPY) )
-    { for(a = table[v]; a; a = a->next)
-      { DEBUG(MSG_HASH_STAT, cmps++);
-	ref = a->references;
-	if ( ATOM_IS_VALID(ref) &&
-	     length == a->length &&
-	     type == a->type &&
-	     memcmp(s, a->name, length) == 0 )
-	{
+  { for(a = table[v]; a; a = a->next)
+    { DEBUG(MSG_HASH_STAT, GD->atoms.cmps++);
+      ref = a->references;
+      if ( ATOM_IS_RESERVED(ref) &&
+	   length == a->length &&
+	   type == a->type &&
+	   same_name(a, s, length, type) )
+      { if ( !ATOM_IS_VALID(ref) )
+	  goto redo;
 #ifdef O_ATOMGC
-	  if ( indexAtom(a->atom) >= GD->atoms.builtin )
-	  {
-#ifdef ATOMIC_REFERENCES
-	  redo_refbump1:
-	    if ( !COMPARE_AND_SWAP(&a->references, ref, ref+1) )
-	    { ref = a->references;
-	      if ( !ATOM_IS_VALID(ref) )
-	        continue;
-	      goto redo_refbump1;
-	    }
-	    if ( ATOM_REF_COUNT(ref+1) == 1 )
-	      ATOMIC_DEC(&GD->atoms.unregistered);
-#else
-	    if ( ATOM_REF_COUNT(++a->references) == 1 )
-	      GD->atoms.unregistered--;
+        if ( indexAtom(a->atom) >= GD->atoms.builtin &&
+	     !likely(bump_atom_references(a, ref)) )
+	  break;			/* atom was just GCed.  Re-create */
 #endif
-	  }
-#endif
-	  *new = FALSE;
-	  release_atom_table();
-	  release_atom_bucket();
-	  return a->atom;
-	}
-      }
-    } else
-    { for(a = table[v]; a; a = a->next)
-      { DEBUG(MSG_HASH_STAT, cmps++);
-	ref = a->references;
-
-	if ( ATOM_IS_VALID(ref) &&
-	     length == a->length &&
-	     type == a->type &&
-	     s == a->name )
-	{
-#ifdef O_ATOMGC
-#ifdef ATOMIC_REFERENCES
-	redo_refbump2:
-	  if ( !COMPARE_AND_SWAP(&a->references, ref, ref+1) )
-	  { ref = a->references;
-	    if ( !ATOM_IS_VALID(ref) )
-	      continue;
-	    goto redo_refbump2;
-	  }
-	  if ( ATOM_REF_COUNT(ref+1) == 1 )
-	    ATOMIC_DEC(&GD->atoms.unregistered);
-#else
-	  if ( ATOM_REF_COUNT(a->references++) == 0 )
-	    GD->atoms.unregistered--;
-#endif
-#endif
-	  *new = FALSE;
-	  release_atom_table();
-	  release_atom_bucket();
-	  return a->atom;
-	}
+        *new = FALSE;
+	release_atom_table();
+	release_atom_bucket();
+	return a->atom;
       }
     }
   }
 
-  if ( atomTable->buckets * 2 < GD->statistics.atoms )
-  { LOCK();
-    rehashAtoms();
-    UNLOCK();
+  if ( GD->atoms.table->buckets * 2 < GD->statistics.atoms )
+  { int rc;
+
+    PL_LOCK(L_REHASH_ATOMS);
+    rc = rehashAtoms();
+    PL_UNLOCK(L_REHASH_ATOMS);
+
+    if ( !rc )
+      outOfCore();
   }
 
-  if ( !( table == atomTable->table && head == table[v] ) )
+  if ( !( table == GD->atoms.table->table && head == table[v] ) )
     goto redo;
 
   a = reserveAtom();
@@ -621,9 +598,13 @@ redo:
   if ( true(type, PL_BLOB_UNIQUE) )
   { a->next = table[v];
     if ( !( COMPARE_AND_SWAP(&table[v], head, a) &&
-            table == atomTable->table ) )
+	    !GD->atoms.rehashing &&	/* See (**) above */
+            table == GD->atoms.table->table ) )
     { if ( false(type, PL_BLOB_NOCOPY) )
         PL_free(a->name);
+      a->type = ATOM_TYPE_INVALID;
+      a->name = "<race>";
+      MemoryBarrier();
       a->references = 0;
       goto redo;
     }
@@ -633,16 +614,6 @@ redo:
   a->references = 1 | ATOM_VALID_REFERENCE | ATOM_RESERVED_REFERENCE;
 #endif
 
-  ATOMIC_INC(&GD->statistics.atoms);
-
-#ifdef O_ATOMGC
-  if ( GD->atoms.margin != 0 &&
-       GD->atoms.unregistered >= GD->atoms.non_garbage + GD->atoms.margin )
-  { if ( GD->statistics.atoms % 128 == 0 ) /* see (*) above */
-      PL_raise(SIG_ATOM_GC);
-  }
-#endif
-
   *new = TRUE;
   if ( type->acquire )
     (*type->acquire)(a->atom);
@@ -650,6 +621,9 @@ redo:
   release_atom_table();
   release_atom_bucket();
 
+  if ( ATOMIC_INC(&GD->statistics.atoms) % 128 == 0 )
+    considerAGC();
+
   return a->atom;
 }
 
@@ -787,11 +761,7 @@ markAtom(atom_t a)
     if ( atomLogFd )
       Sfprintf(atomLogFd, "Marked `%s' at (#%d)\n", ap->name, i);
 #endif
-#ifdef ATOMIC_REFERENCES
     ATOMIC_OR(&ap->references, ATOM_MARKED_REFERENCE);
-#else
-    ap->references |= ATOM_MARKED_REFERENCE;
-#endif
   }
 }
 
@@ -802,10 +772,11 @@ unmarkAtoms(void)
 
   for(index=GD->atoms.builtin, i=MSB(index); !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
@@ -813,12 +784,7 @@ unmarkAtoms(void)
     { Atom a = b + index;
 
       if ( ATOM_IS_MARKED(a->references) )
-      {
-#ifdef ATOMIC_REFERENCES
-        ATOMIC_AND(&a->references, ~ATOM_MARKED_REFERENCE);
-#else
-        a->references &= ~ATOM_MARKED_REFERENCE;
-#endif
+      { ATOMIC_AND(&a->references, ~ATOM_MARKED_REFERENCE);
       }
     }
   }
@@ -828,7 +794,7 @@ unmarkAtoms(void)
 void
 maybe_free_atom_tables(void)
 {
-  AtomTable t = atomTable;
+  AtomTable t = GD->atoms.table;
   while ( t )
   { AtomTable t2 = t->prev;
     if ( t2 && !pl_atom_table_in_use(t2) )
@@ -856,22 +822,27 @@ static int
 invalidateAtom(Atom a, unsigned int ref)
 { Atom *ap;
 
-  if ( !COMPARE_AND_SWAP(&a->references, ref, (ref & ~ATOM_VALID_REFERENCE)) )
+#define ATOM_PRE_DESTROY_REFERENCE \
+	(ATOM_DESTROY_REFERENCE|ATOM_RESERVED_REFERENCE)
+
+  if ( !COMPARE_AND_SWAP(&a->references, ref, ATOM_PRE_DESTROY_REFERENCE) )
   { return FALSE;
   }
 
   if ( a->type->release )
   { if ( !(*a->type->release)(a->atom) )
-    { ATOMIC_OR(&a->references, ATOM_VALID_REFERENCE);
+    { COMPARE_AND_SWAP(&a->references, ATOM_PRE_DESTROY_REFERENCE, ref);
       return FALSE;
     }
   } else if ( GD->atoms.gc_hook )
   { if ( !(*GD->atoms.gc_hook)(a->atom) )
-    { ATOMIC_OR(&a->references, ATOM_VALID_REFERENCE);
+    { COMPARE_AND_SWAP(&a->references, ATOM_PRE_DESTROY_REFERENCE, ref);
       return FALSE;				/* foreign hooks says `no' */
     }
   }
 
+  a->references = ATOM_DESTROY_REFERENCE;
+
 #ifdef O_DEBUG_ATOMGC
   if ( atomLogFd )
     Sfprintf(atomLogFd, "Invalidated `%s'\n", a->name);
@@ -882,7 +853,7 @@ invalidateAtom(Atom a, unsigned int ref)
     uintptr_t mask;
 
   redo:
-    table = atomTable;
+    table = GD->atoms.table;
     mask = table->buckets-1;
     ap = &table->table[a->hash_value & mask];
 
@@ -924,7 +895,7 @@ destroyAtom(Atom a, Atom **buckets)
   size_t index;
 
   while ( buckets && *buckets )
-  { t = atomTable;
+  { t = GD->atoms.table;
     while ( t )
     { v = a->hash_value & (t->buckets-1);
       if ( *buckets == t->table+v )
@@ -951,8 +922,9 @@ destroyAtom(Atom a, Atom **buckets)
   { PL_free(a->name);
   }
 
-  a->name = NULL;
-  a->type = NULL;
+  a->name = "<reclaimed>";
+  a->type = ATOM_TYPE_INVALID;
+  MemoryBarrier();
   a->references = 0;
 
   index = indexAtom(a->atom);
@@ -973,10 +945,11 @@ collectAtoms(void)
 
   for(index=GD->atoms.builtin, i=MSB(index); !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
@@ -991,12 +964,7 @@ collectAtoms(void)
       if ( !ATOM_IS_MARKED(ref) && (ATOM_REF_COUNT(ref) == 0) )
       { invalidateAtom(a, ref);
       } else
-      {
-#ifdef ATOMIC_REFERENCES
-	ATOMIC_AND(&a->references, ~ATOM_MARKED_REFERENCE);
-#else
-	a->references &= ~ATOM_MARKED_REFERENCE;
-#endif
+      {	ATOMIC_AND(&a->references, ~ATOM_MARKED_REFERENCE);
         if ( ATOM_REF_COUNT(ref) == 0 )
 	  unregistered++;
       }
@@ -1051,44 +1019,24 @@ pl_garbage_collect_atoms(void)
   double t;
   sigset_t set;
   size_t reclaimed;
+  int rc = TRUE;
 
   if ( GD->cleaning != CLN_NORMAL )	/* Cleaning up */
     return TRUE;
 
-  PL_LOCK(L_GC);
-  if ( gc_status.blocked )		/* Tricky things; avoid problems. */
-  { PL_UNLOCK(L_GC);
-    succeed;
-  }
-
-#ifdef O_PLMT
-  if ( GD->gc.active )			/* GC in progress: delay */
-  { DEBUG(MSG_AGC, Sdprintf("GC active; delaying AGC\n"));
-    GD->gc.agc_waiting = TRUE;
-    PL_UNLOCK(L_GC);
-    succeed;
-  }
-#endif
-
-  gc_status.blocked++;			/* avoid recursion */
+  if ( !COMPARE_AND_SWAP(&GD->atoms.gc_active, FALSE, TRUE) )
+    return TRUE;
 
   if ( verbose )
-  {
-#ifdef O_DEBUG_ATOMGC
-/*
-    Sdprintf("Starting ATOM-GC.  Stack:\n");
-    PL_backtrace(5, 0);
-*/
-#endif
-    printMessage(ATOM_informational,
-		 PL_FUNCTOR_CHARS, "agc", 1,
-		   PL_CHARS, "start");
+  { if ( !printMessage(ATOM_informational,
+		       PL_FUNCTOR_CHARS, "agc", 1,
+		         PL_CHARS, "start") )
+    { GD->atoms.gc_active = FALSE;
+      return FALSE;
+    }
   }
 
-  PL_LOCK(L_THREAD);
-  PL_LOCK(L_AGC);
-  LOCK();
-  GD->atoms.gc_active = TRUE;
+  PL_LOCK(L_REHASH_ATOMS);
   blockSignals(&set);
   t = CpuTime(CPU_USER);
   unmarkAtoms();
@@ -1099,28 +1047,25 @@ pl_garbage_collect_atoms(void)
 #endif
   oldcollected = GD->atoms.collected;
   reclaimed = collectAtoms();
-  GD->atoms.gc_active = FALSE;
   GD->atoms.collected += reclaimed;
   ATOMIC_SUB(&GD->statistics.atoms, reclaimed);
   t = CpuTime(CPU_USER) - t;
   GD->atoms.gc_time += t;
   GD->atoms.gc++;
   unblockSignals(&set);
-  UNLOCK();
-  PL_UNLOCK(L_AGC);
-  PL_UNLOCK(L_THREAD);
-  gc_status.blocked--;
-  PL_UNLOCK(L_GC);
+  PL_UNLOCK(L_REHASH_ATOMS);
 
   if ( verbose )
-    printMessage(ATOM_informational,
-		 PL_FUNCTOR_CHARS, "agc", 1,
-		   PL_FUNCTOR_CHARS, "done", 3,
-		     PL_INT64, GD->atoms.collected - oldcollected,
-		     PL_INT, GD->statistics.atoms,
-		     PL_DOUBLE, (double)t);
+    rc = printMessage(ATOM_informational,
+		      PL_FUNCTOR_CHARS, "agc", 1,
+		        PL_FUNCTOR_CHARS, "done", 3,
+		          PL_INT64, GD->atoms.collected - oldcollected,
+		          PL_INT, GD->statistics.atoms,
+		          PL_DOUBLE, (double)t);
 
-  succeed;
+  GD->atoms.gc_active = FALSE;
+
+  return rc;
 }
 
 
@@ -1133,6 +1078,15 @@ PL_agc_hook(PL_agc_hook_t new)
 }
 
 
+static void
+considerAGC(void)
+{ if ( GD->atoms.margin != 0 &&
+       GD->atoms.unregistered >= GD->atoms.non_garbage + GD->atoms.margin )
+  { signalGCThread(SIG_ATOM_GC);
+  }
+}
+
+
 #endif /*O_ATOMGC*/
 
 #undef PL_register_atom
@@ -1153,17 +1107,21 @@ operations. This should be safe because:
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-register_atom(Atom p)
-{
-#ifdef ATOMIC_REFERENCES
-  if ( (ATOMIC_INC(&p->references) & ATOM_REF_COUNT_MASK) == 1 )
-    ATOMIC_DEC(&GD->atoms.unregistered);
-#else
-  LOCK();
-  if ( p->references++ == 0 )
-    GD->atoms.unregistered--;
-  UNLOCK();
-#endif
+register_atom(volatile Atom p)
+{ for(;;)
+  { unsigned int ref  = p->references;
+    unsigned int nref = ref+1;
+
+    if ( ATOM_REF_COUNT(nref) != 0 )
+    { if ( COMPARE_AND_SWAP(&p->references, ref, nref) )
+      { if ( ATOM_REF_COUNT(nref) == 1 )
+	  ATOMIC_DEC(&GD->atoms.unregistered);
+	return;
+      }
+    } else
+    { return;
+    }
+  }
 }
 
 
@@ -1182,6 +1140,23 @@ PL_register_atom(atom_t a)
 }
 
 
+static char *
+dbgAtomName(Atom a, char *enc, char **buf)
+{ if ( a->type == &text_atom )
+  { if ( enc ) *enc = 'L';
+    return a->name;
+  } else if ( isUCSAtom(a) )
+  { if ( enc ) *enc = 'W';
+    return a->name;
+  } else
+  { size_t len = 0;
+    IOSTREAM *fd = Sopenmem(buf, &len, "w");
+    (a->type->write)(fd, a->atom, 0);
+    Sclose(fd);
+    return *buf;
+  }
+}
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Foreign code reduces the reference count. This is safe, unless we are in
 the following scenario:
@@ -1232,15 +1207,20 @@ PL_unregister_atom(atom_t a)
       trap_gdb();
     }
 
-#ifdef ATOMIC_REFERENCES
+    if ( unlikely(ATOM_REF_COUNT(p->references+1) == 0) )
+      return;
+
     if ( GD->atoms.gc_active )
     { unsigned int oldref, newref;
 
       do
       { oldref = p->references;
-        newref = (ATOM_REF_COUNT(oldref) == 1) ? ((oldref-1) | ATOM_MARKED_REFERENCE) : oldref-1;
+	newref = oldref - 1;
+
+	if ( ATOM_REF_COUNT(newref) == 0 )
+	  newref |= ATOM_MARKED_REFERENCE;
       } while( !COMPARE_AND_SWAP(&p->references, oldref, newref) );
-      refs = newref;
+      refs = ATOM_REF_COUNT(newref);
     } else
     { GET_LD
 
@@ -1249,41 +1229,72 @@ PL_unregister_atom(atom_t a)
       if ( (refs=ATOM_REF_COUNT(ATOMIC_DEC(&p->references))) == 0 )
 	ATOMIC_INC(&GD->atoms.unregistered);
     }
-#else
-    LOCK();
-    if ( (refs=ATOM_REF_COUNT(--p->references)) == 0 )
-      GD->atoms.unregistered++;
-    UNLOCK();
-#endif
-    if ( refs == (unsigned int)-1 )
-    { Sdprintf("OOPS: PL_unregister_atom('%s'): -1 references\n", p->name);
+    if ( refs == ATOM_REF_COUNT((unsigned int)-1) )
+    { char fmt[100];
+      char *enc;
+      char *buf = NULL;
+
+      strcpy(fmt, "OOPS: PL_unregister_atom('%Ls'): -1 references\n");
+      enc = strchr(fmt, '%')+1;
+
+      Sdprintf(fmt, dbgAtomName(p, enc, &buf));
+      if ( buf )
+	PL_free(buf);
       trap_gdb();
     }
   }
 #endif
 }
 
-#define PL_register_atom error		/* prevent using them after this */
-#define PL_unregister_atom error
-
 
 		 /*******************************
 		 *	      CHECK		*
 		 *******************************/
-#if O_DEBUG
+
+#ifdef O_DEBUG
+
+static int
+findAtomSelf(Atom a)
+{ GET_LD
+  Atom *table;
+  int buckets;
+  Atom head, ap;
+  unsigned int v;
+
+redo:
+  acquire_atom_table(table, buckets);
+  v = a->hash_value & (buckets-1);
+  head = table[v];
+  acquire_atom_bucket(table+v);
+
+  for(ap=head; ap; ap = ap->next )
+  { if ( ap == a )
+    { release_atom_table();
+      release_atom_bucket();
+      return TRUE;
+    }
+  }
+
+  if ( !( table == GD->atoms.table->table && head == table[v] ) )
+    goto redo;
+
+  return FALSE;
+}
+
 
 int
-checkAtoms(void)
+checkAtoms_src(const char *file, int line)
 { size_t index;
   int i, last=FALSE;
   int errors = 0;
 
   for(index=1, i=0; !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
@@ -1293,16 +1304,23 @@ checkAtoms(void)
       if ( ATOM_IS_VALID(a->references) )
       { if ( !a->type || !a->name || (int)ATOM_REF_COUNT(a->references) < 0 )
 	{ size_t bs = (size_t)1<<i;
-	  Sdprintf("Invalid atom %p at index %zd in block at %p (size %d)\n",
-		   a, index, b+bs, bs);
+	  Sdprintf("%s%d: invalid atom %p at index %zd in "
+		   "block at %p (size %d)\n",
+		   file, line, a, index, b+bs, bs);
 	  errors++;
 	  trap_gdb();
 	}
+
+	if ( true(a->type, PL_BLOB_UNIQUE) )
+	{ if ( !findAtomSelf(a) )
+	  { Sdprintf("%s%d: cannot find self: %p\n", file, line, a);
+	  }
+	}
       }
     }
   }
 
-  return errors == 0;
+  return errors;
 }
 
 #endif /*O_DEBUG*/
@@ -1312,7 +1330,7 @@ checkAtoms(void)
 		 *	    REHASH TABLE	*
 		 *******************************/
 
-static void
+static int
 rehashAtoms(void)
 { AtomTable newtab;
   uintptr_t mask;
@@ -1320,44 +1338,61 @@ rehashAtoms(void)
   int i, last=FALSE;
 
   if ( GD->cleaning != CLN_NORMAL )
-    return;				/* no point anymore and foreign ->type */
+    return TRUE;			/* no point anymore and foreign ->type */
 					/* pointers may have gone */
 
-  if ( atomTable->buckets * 2 >= GD->statistics.atoms )
-    return;
+  if ( GD->atoms.table->buckets * 2 >= GD->statistics.atoms )
+    return TRUE;
 
-  newtab = allocHeapOrHalt(sizeof(*newtab));
-  newtab->buckets = atomTable->buckets * 2;
-  newtab->table = allocHeapOrHalt(newtab->buckets * sizeof(Atom));
+  if ( !(newtab = allocHeap(sizeof(*newtab))) )
+    return FALSE;
+  newtab->buckets = GD->atoms.table->buckets * 2;
+  if ( !(newtab->table = allocHeapOrHalt(newtab->buckets * sizeof(Atom))) )
+  { freeHeap(newtab, sizeof(*newtab));
+    return FALSE;
+  }
   memset(newtab->table, 0, newtab->buckets * sizeof(Atom));
-  newtab->prev = atomTable;
+  newtab->prev = GD->atoms.table;
   mask = newtab->buckets-1;
 
   DEBUG(MSG_HASH_STAT,
-	Sdprintf("rehashing atoms (%d --> %d)\n", atomTable->buckets, newtab->buckets));
+	Sdprintf("rehashing atoms (%d --> %d)\n",
+		 GD->atoms.table->buckets, newtab->buckets));
+
+  GD->atoms.rehashing = TRUE;
 
   for(index=1, i=0; !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
     for(; index<upto; index++)
-    { Atom a = b + index;
-
-      if ( ATOM_IS_VALID(a->references) && true(a->type, PL_BLOB_UNIQUE) )
-      { size_t v = a->hash_value & mask;
-
-	a->next = newtab->table[v];
-	newtab->table[v] = a;
+    { volatile Atom a = b + index;
+      unsigned int ref;
+redo:
+      ref = a->references;
+      if ( ATOM_IS_RESERVED(ref) )
+      { if ( !ATOM_IS_VALID(ref) )
+	  goto redo;
+	if ( true(a->type, PL_BLOB_UNIQUE) )
+	{ size_t v;
+	  v = a->hash_value & mask;
+	  a->next = newtab->table[v];
+	  newtab->table[v] = a;
+	}
       }
     }
   }
 
-  atomTable = newtab;
+  GD->atoms.table = newtab;
+  GD->atoms.rehashing = FALSE;
+
+  return TRUE;
 }
 
 
@@ -1395,7 +1430,7 @@ resetListAtoms(void)
 { Atom a = atomValue(ATOM_dot);
 
   if ( strcmp(a->name, ".") != 0 )
-  { Atom *ap2 = &atomTable->table[a->hash_value & (atomTable->buckets-1)];
+  { Atom *ap2 = &GD->atoms.table->table[a->hash_value & (GD->atoms.table->buckets-1)];
     unsigned int v;
     static char *s = ".";
 
@@ -1415,10 +1450,10 @@ resetListAtoms(void)
     a->name   = s;
     a->length = strlen(s);
     a->hash_value = MurmurHashAligned2(s, a->length, MURMUR_SEED);
-    v = a->hash_value & (atomTable->buckets-1);
+    v = a->hash_value & (GD->atoms.table->buckets-1);
 
-    a->next      = atomTable->table[v];
-    atomTable->table[v] = a;
+    a->next      = GD->atoms.table->table[v];
+    GD->atoms.table->table[v] = a;
   }
 
   a = atomValue(ATOM_nil);
@@ -1456,7 +1491,7 @@ registerBuiltinAtoms(void)
     }
 
     v0 = MurmurHashAligned2(s, len, MURMUR_SEED);
-    v  = v0 & (atomTable->buckets-1);
+    v  = v0 & (GD->atoms.table->buckets-1);
 
     a = &GD->atoms.array.blocks[idx][index];
     a->atom       = (index<<LMASK_BITS)|TAG_ATOM;
@@ -1469,8 +1504,8 @@ registerBuiltinAtoms(void)
 #ifdef O_TERMHASH
     a->hash_value = v0;
 #endif
-    a->next       = atomTable->table[v];
-    atomTable->table[v]  = a;
+    a->next       = GD->atoms.table->table[v];
+    GD->atoms.table->table[v]  = a;
 
     GD->atoms.no_hole_before = index+1;
     GD->atoms.highest = index+1;
@@ -1485,7 +1520,7 @@ exitAtoms(int status, void *context)
   (void)context;
 
   Sdprintf("hashstat: %d lookupAtom() calls used %d strcmp() calls\n",
-	   lookups, cmps);
+	   GD->atoms.lookups, GD->atoms.cmps);
 
   return 0;
 }
@@ -1493,14 +1528,14 @@ exitAtoms(int status, void *context)
 
 
 void
-initAtoms(void)
-{ LOCK();
-  if ( !atomTable )			/* Atom hash table */
-  { atomTable = allocHeapOrHalt(sizeof(*atomTable));
-    atomTable->buckets = ATOMHASHSIZE;
-    atomTable->table = allocHeapOrHalt(ATOMHASHSIZE * sizeof(Atom));
-    memset(atomTable->table, 0, ATOMHASHSIZE * sizeof(Atom));
-    atomTable->prev = NULL;
+do_init_atoms(void)
+{ PL_LOCK(L_INIT_ATOMS);
+  if ( !GD->atoms.initialised )			/* Atom hash table */
+  { GD->atoms.table = allocHeapOrHalt(sizeof(*GD->atoms.table));
+    GD->atoms.table->buckets = ATOMHASHSIZE;
+    GD->atoms.table->table = allocHeapOrHalt(ATOMHASHSIZE * sizeof(Atom));
+    memset(GD->atoms.table->table, 0, ATOMHASHSIZE * sizeof(Atom));
+    GD->atoms.table->prev = NULL;
 
     GD->atoms.highest = 1;
     GD->atoms.no_hole_before = 1;
@@ -1516,10 +1551,9 @@ initAtoms(void)
 #ifdef O_RESERVED_SYMBOLS
     initReservedSymbols();
 #endif
-
-
+    GD->atoms.initialised = TRUE;
   }
-  UNLOCK();
+  PL_UNLOCK(L_INIT_ATOMS);
 }
 
 
@@ -1537,10 +1571,11 @@ cleanupAtoms(void)
 
   for(index=GD->atoms.builtin, i=MSB(index); !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
@@ -1576,15 +1611,15 @@ cleanupAtoms(void)
     }
   }
 
-  table = atomTable;
+  table = GD->atoms.table;
   while ( table )
   { AtomTable prev = table->prev;
     freeHeap(table->table, table->buckets * sizeof(Atom));
     freeHeap(table, sizeof(atom_table));
     table = prev;
   }
-  if ( atomTable )
-  { atomTable = NULL;
+  if ( GD->atoms.table )
+  { GD->atoms.table = NULL;
   }
 }
 
@@ -1603,7 +1638,7 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t state ARG_LD)
       { if ( type )
 	  return PL_unify_atom(type, bt->atom_name);
 	else if ( false(bt, PL_BLOB_TEXT) )
-	  fail;
+	  return FALSE;
 
 	succeed;
       }
@@ -1618,45 +1653,61 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t state ARG_LD)
       break;
     case FRG_CUTTED:
     default:
-      succeed;
+      return TRUE;
   }
 
   if ( type )
   { if ( !PL_is_variable(type) &&
 	 !PL_get_atom_ex(type, &type_name) )
-      fail;
+      return FALSE;
   }
 
-  PL_LOCK(L_AGC);
   for(i=MSB(index); !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
     for(; index<upto; index++)
     { Atom atom = b + index;
+      unsigned int refs = atom->references;
+      PL_blob_t *btype = atom->type;
+      int rc;
 
-      if ( ATOM_IS_VALID(atom->references) &&
-	   atom->atom != ATOM_garbage_collected )
-      { if ( type )
-	{ if ( type_name && type_name != atom->type->atom_name )
-	    continue;
+      if ( ATOM_IS_VALID(refs) && btype &&
+	   (!type_name || type_name == btype->atom_name) &&
+	   bump_atom_references(atom, refs) )
+      { DEBUG(CHK_SECURE,	/* avoid trap through linkVal__LD() check */
+	      if ( atom->atom == ATOM_garbage_collected )
+	      { PL_unregister_atom(atom->atom);
+	        continue;
+	      });
 
-	  PL_unify_atom(type, atom->type->atom_name);
-	} else if ( false(atom->type, PL_BLOB_TEXT) )
+	if ( type )
+	{ if ( !type_name )
+	  { if ( !PL_unify_atom(type, btype->atom_name) )
+	    { PL_unregister_atom(atom->atom);
+	      return FALSE;
+	    }
+	  }
+	} else if ( false(btype, PL_BLOB_TEXT) )
+	{ PL_unregister_atom(atom->atom);
 	  continue;
+	}
 
-	PL_unify_atom(a, atom->atom);
-	PL_UNLOCK(L_AGC);
-	ForeignRedoInt(index+1);
+	rc = PL_unify_atom(a, atom->atom);
+	PL_unregister_atom(atom->atom);
+	if ( rc )
+	  ForeignRedoInt(index+1);
+	else
+	  return rc;
       }
     }
   }
-  PL_UNLOCK(L_AGC);
 
   return FALSE;
 }
@@ -1723,12 +1774,6 @@ typedef struct match
 } *Match;
 
 
-static inline int
-completion_candidate(Atom a)
-{ return (a->references || indexAtom(a->atom) < GD->atoms.builtin);
-}
-
-
 /* An atom without references cannot be part of the program
 */
 
@@ -1789,10 +1834,11 @@ extendAtom(char *prefix, bool *unique, char *common)
 
   for(index=1, i=0; !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
@@ -1801,7 +1847,6 @@ extendAtom(char *prefix, bool *unique, char *common)
 
       if ( ATOM_IS_VALID(a->references) && a->type == &text_atom &&
 	   global_atom(a) &&
-	   completion_candidate(a) &&
 	   strprefix(a->name, prefix) &&
 	   strlen(a->name) < LINESIZ )
       { if ( first == TRUE )
@@ -1880,10 +1925,11 @@ extend_alternatives(PL_chars_t *prefix, struct match *altv, int *altn)
   *altn = 0;
   for(index=1, i=0; !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
@@ -1896,7 +1942,6 @@ extend_alternatives(PL_chars_t *prefix, struct match *altv, int *altn)
 
       if ( ATOM_IS_VALID(a->references) &&
 	   global_atom(a) &&
-	   completion_candidate(a) &&
 	   get_atom_ptr_text(a, &hit) &&
 	   hit.length < ALT_SIZ &&
 	   PL_cmp_text(prefix, 0, &hit, 0, prefix->length) == 0 &&
@@ -2003,10 +2048,11 @@ atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
 
   for(i=MSB(index); !last; i++)
   { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
     Atom b = GD->atoms.array.blocks[i];
 
-    if ( upto >= GD->atoms.highest )
-    { upto = GD->atoms.highest;
+    if ( upto >= high )
+    { upto = high;
       last = TRUE;
     }
 
@@ -2014,9 +2060,11 @@ atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
     { Atom a = b + index;
 
       if ( is_signalled(PASS_LD1) )	/* Notably allow windows version */
-	PL_handle_signals();		/* to break out on ^C */
+      { if ( PL_handle_signals() < 0 )	/* to break out on ^C */
+	  return FALSE;
+      }
 
-      if ( ATOM_IS_VALID(a->references) && completion_candidate(a) &&
+      if ( ATOM_IS_VALID(a->references) && global_atom(a) &&
 	   get_atom_ptr_text(a, hit) &&
 	   hit->length < ALT_SIZ &&
 	   PL_cmp_text(prefix, 0, hit, 0, prefix->length) == 0 &&
