@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2015, University of Amsterdam,
+    Copyright (c)  1985-2017, University of Amsterdam,
                               VU University Amsterdam
     All rights reserved.
 
@@ -776,6 +776,7 @@ typedef enum
 #define GF_PROCEDURE	0x2		/* check for max arity */
 #define GF_NAMEARITY	0x4		/* only accept name/arity */
 
+#define SM_NOCREATE	0x1		/* stripModule(): do not create modules */
 
 		 /*******************************
 		 *	       ALERT		*
@@ -823,8 +824,8 @@ with one operation, it turns out to be faster as well.
 
 #define true(s, a)		((s)->flags & (a))
 #define false(s, a)		(!true((s), (a)))
-#define set(s, a)		((s)->flags |= (a))
-#define clear(s, a)		((s)->flags &= ~(a))
+#define set(s, a)		ATOMIC_OR(&(s)->flags, (a))
+#define clear(s, a)		ATOMIC_AND(&(s)->flags, ~(a))
 #define clearFlags(s)		((s)->flags = 0)
 
 /* Flags on predicates (packed in unsigned int */
@@ -929,8 +930,7 @@ Macros for environment frames (local stack frames)
 #define FR_INBOX		(0x0040) /* Inside box (for REDO in built-in) */
 #define FR_CONTEXT		(0x0080) /* fr->context is set */
 #define FR_CLEANUP		(0x0100) /* setup_call_cleanup/4: marked for cleanup */
-#define FR_KEEPLTOP		(0x0200) /* Continuations: to not reset lTop */
-#define FR_INRESET		(0x0400) /* Continuations: inside reset/3 */
+#define FR_INRESET		(0x0200) /* Continuations: inside reset/3 */
 #define FR_MAGIC_MASK		(0xfffff000)
 #define FR_MAGIC_MASK2		(0xffff0000)
 #define FR_MAGIC		(0x549d5000)
@@ -1014,7 +1014,7 @@ typedef struct ggen_t
 #endif /*O_LOGICAL_UPDATE*/
 
 #define FR_CLEAR_NEXT	FR_SKIPPED|FR_WATCHED|FR_CATCHED|FR_HIDE_CHILDS|FR_CLEANUP
-#define FR_CLEAR_FLAGS	(FR_CLEAR_NEXT|FR_CONTEXT|FR_KEEPLTOP)
+#define FR_CLEAR_FLAGS	(FR_CLEAR_NEXT|FR_CONTEXT)
 
 #define setNextFrameFlags(next, fr) \
 	do \
@@ -1136,19 +1136,23 @@ typedef struct atom_table
 
 #ifdef O_ATOMGC
 
-#define ATOM_STATE_MASK		((unsigned int)0x7 << (INTBITSIZE-3))
+#define ATOM_STATE_MASK		((unsigned int)0xF << (INTBITSIZE-4))
 #define ATOM_RESERVED_REFERENCE	((unsigned int)0x1 << (INTBITSIZE-1))
 #define ATOM_VALID_REFERENCE	((unsigned int)0x1 << (INTBITSIZE-2))
 #define ATOM_MARKED_REFERENCE	((unsigned int)0x1 << (INTBITSIZE-3))
+#define ATOM_DESTROY_REFERENCE	((unsigned int)0x1 << (INTBITSIZE-4))
 
 #define ATOM_IS_FREE(ref)	(((ref) & ATOM_STATE_MASK) == 0)
 #define ATOM_IS_RESERVED(ref)	((ref) & ATOM_RESERVED_REFERENCE)
 #define ATOM_IS_VALID(ref)	((ref) & ATOM_VALID_REFERENCE)
 #define ATOM_IS_MARKED(ref)	((ref) & ATOM_MARKED_REFERENCE)
+#define ATOM_IS_DESTROYED(ref)	((ref) & ATOM_DESTROY_REFERENCE)
 
 #define ATOM_REF_COUNT_MASK	(~ATOM_STATE_MASK)
 #define ATOM_REF_COUNT(ref)	((ref) & ATOM_REF_COUNT_MASK)
 
+#define ATOM_TYPE_INVALID	((PL_blob_t*)0x007)
+
 #ifdef O_DEBUG_ATOMGC
 #define PL_register_atom(a) \
 	_PL_debug_register_atom(a, __FILE__, __LINE__, __PRETTY_FUNCTION__)
@@ -1191,14 +1195,27 @@ typedef struct functor_table
 
 
 #ifdef O_LOGICAL_UPDATE
-#define visibleClause(cl, gen) \
-	((cl)->generation.created <= (gen) && \
-	 (cl)->generation.erased   > (gen))
+#define VISIBLE_CLAUSE(cl, gen) \
+	( ( (cl)->generation.created <= (gen) && \
+	    (cl)->generation.erased   > (gen) && \
+	    (cl)->generation.erased  != LD->gen_reload \
+	  ) || \
+	  ( (cl)->generation.created == LD->gen_reload \
+	  ) \
+	)
+#define GLOBALLY_VISIBLE_CLAUSE(cl, gen) \
+	( (cl)->generation.created <= (gen) && \
+	  (cl)->generation.erased   > (gen) \
+	)
 #else
-#define visibleClause(cl, gen) false(cl, CL_ERASED)
+#define VISIBLE_CLAUSE(cl, gen) false(cl, CL_ERASED)
+#define GLOBALLY_VISIBLE_CLAUSE(cl, gen) false(cl, CL_ERASED)
 #endif
 
-#define visibleClauseCNT(cl, gen) visibleClause__LD(cl, gen PASS_LD)
+#define visibleClause(cl, gen) visibleClause__LD(cl, gen PASS_LD)
+#define visibleClauseCNT(cl, gen) visibleClauseCNT__LD(cl, gen PASS_LD)
+
+#define GEN_INVALID 0
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Struct clause must be a  multiple   of  sizeof(word)  for compilation on
@@ -1218,8 +1235,8 @@ struct clause
 #endif /*O_LOGICAL_UPDATE*/
   unsigned int		variables;	/* # of variables for frame */
   unsigned int		prolog_vars;	/* # real Prolog variables */
-  unsigned		flags : 8;	/* Flag field holding: */
-  unsigned		line_no : 24;	/* Source line-number */
+  unsigned int		flags;		/* Flag field holding: */
+  unsigned int		line_no;	/* Source line-number */
   unsigned int		source_no;	/* Index of source-file */
   unsigned int		owner_no;	/* Index of owning source-file */
   unsigned int		references;	/* # ClauseRef pointing at me */
@@ -1230,10 +1247,11 @@ struct clause
 typedef struct clause_list
 { ClauseRef	first_clause;		/* clause list of procedure */
   ClauseRef	last_clause;		/* last clause of list */
-  ClauseIndex	clause_indexes;		/* Hash index(es) */
+  ClauseIndex  *clause_indexes;		/* Hash index(es) */
   unsigned int	number_of_clauses;	/* number of associated clauses */
   unsigned int	erased_clauses;		/* number of erased clauses in set */
   unsigned int	number_of_rules;	/* number of real rules */
+  unsigned int	jiti_tried;		/* number of times we tried to find */
 } clause_list, *ClauseList;
 
 typedef struct clause_ref
@@ -1319,7 +1337,7 @@ struct clause_bucket
   unsigned int	dirty;			/* # of garbage clauses */
 };
 
-#define MAX_MULTI_INDEX 1
+#define MAX_MULTI_INDEX 4
 
 struct clause_index
 { unsigned int	 buckets;		/* # entries */
@@ -1329,17 +1347,16 @@ struct clause_index
   unsigned int	 dirty;			/* # chains that are dirty */
   unsigned short args[MAX_MULTI_INDEX];	/* Indexed arguments */
   unsigned	 is_list : 1;		/* Index with lists */
-  unsigned	 incomplete : 1;	/* Not all clauses are in the index */
   float		 speedup;		/* Estimated speedup */
-  struct bit_vector *tried_better;	/* We tried to access for better hash */
-  ClauseIndex	 next;			/* Next index */
   ClauseBucket	 entries;		/* chains holding the clauses */
 };
 
-typedef struct clause_index_list
-{ ClauseIndex index;
-  struct clause_index_list *next;
-} clause_index_list, *ClauseIndexList;
+typedef struct arg_info
+{ float		speedup;		/* Computed speedup */
+  unsigned	ln_buckets : 5;		/* lg2(bucket count) */
+  unsigned	assessed   : 1;		/* Value was assessed */
+  unsigned	meta	   : 4;		/* Meta-argument info */
+} arg_info;
 
 #define MAX_BLOCKS 20			/* allows for 2M threads */
 
@@ -1348,12 +1365,6 @@ typedef struct local_definitions
   Definition preallocated[7];
 } local_definitions;
 
-#ifdef _MSC_VER
-typedef __int64 meta_mask;		/* MSVC cannot do typedef of typedef!? */
-#else
-typedef uint64_t meta_mask;
-#endif
-
 struct definition
 { FunctorDef	functor;		/* Name/Arity of procedure */
   Module	module;			/* module of the predicate */
@@ -1364,11 +1375,11 @@ struct definition
     Func	function;		/* function pointer of procedure */
     LocalDefinitions local;		/* P_THREAD_LOCAL predicates */
   } impl;
-  ClauseIndexList old_clause_indexes;	/* Outdated hash indexes */
-  struct bit_vector *tried_index;	/* Arguments on which we tried to index */
-  meta_mask	meta_info;		/* meta-predicate info */
+  arg_info     *args;			/* Meta and indexing info */
   unsigned int  flags;			/* booleans (P_*) */
   unsigned int  shared;			/* #procedures sharing this def */
+  struct linger_list  *lingering;	/* Assocated lingering objects */
+  gen_t		last_modified;		/* Generation I was last modified */
 #ifdef O_PROF_PENTIUM
   int		prof_index;		/* index in profiling */
   char	       *prof_name;		/* name in profiling */
@@ -1556,8 +1567,8 @@ struct fliFrame
 struct record
 { int		size;			/* # bytes of the record */
   unsigned      gsize;			/* Size on global stack */
-  unsigned	nvars : 27;		/* # variables in the term */
-  unsigned	flags : 5;		/* Flags, holding */
+  unsigned	nvars;			/* # variables in the term */
+  unsigned	flags;			/* Flags, holding */
 					/* R_ERASED */
 					/* R_EXTERNAL */
 					/* R_DUPLICATE */
@@ -1586,6 +1597,21 @@ struct recordRef
   Record	record;			/* the record itself */
 };
 
+
+		 /*******************************
+		 *	EXCEPTION CLASSES	*
+		 *******************************/
+
+typedef enum except_class
+{ EXCEPT_NONE = 0,			/* no exception */
+  EXCEPT_OTHER,				/* any other exception */
+  EXCEPT_ERROR,				/* ISO error(Formal,Context) */
+  EXCEPT_RESOURCE,			/* ISO error(resource_error(_), _) */
+  EXCEPT_TIMEOUT,			/* time_limit_exceeded */
+  EXCEPT_ABORT				/* '$aborted' */
+} except_class;
+
+
 		 /*******************************
 		 *	SOURCE FILE ADMIN	*
 		 *******************************/
@@ -1597,7 +1623,7 @@ typedef struct p_reload
 { Definition	predicate;		/* definition we are working on */
   gen_t		generation;		/* generation we update */
   ClauseRef	current_clause;		/* currently reloading clause */
-  meta_mask	meta_info;		/* new meta declaration (if any) */
+  arg_info     *args;			/* Meta info on arguments */
   unsigned	flags;			/* new flags (P_DYNAMIC, etc.) */
   unsigned	number_of_clauses;	/* Number of clauses we've seen */
 } p_reload;
@@ -1667,6 +1693,7 @@ struct module
   int		level;		/* Distance to root (root=0) */
   unsigned int	line_no;	/* Source line-number */
   unsigned int  flags;		/* booleans: */
+  gen_t		last_modified;	/* Generation I was last modified */
 };
 
 struct trail_entry
@@ -1689,12 +1716,8 @@ struct gc_trail_entry
 #define MA_HAT		14		/* ^ */
 #define MA_DCG		15		/* // */
 
-#define MA_INFO(def, n) \
-	(((def)->meta_info >> ((n)*4)) & 0xf)
-#define MA_SETINFO(def, n, i) \
-	((def)->meta_info &= ~((meta_mask)0xf << (n)*4), \
-	 (def)->meta_info |= ((meta_mask)(i) << (n)*4))
-
+#define MA_NEEDS_TRANSPARENT(m) \
+	((m) < 10 || (m) == MA_META || (m) == MA_HAT || (m) == MA_DCG)
 
 		 /*******************************
 		 *	     MARK/UNDO		*
@@ -1724,14 +1747,20 @@ struct gc_trail_entry
 
 #define Mark(b)		do { (b).trailtop  = tTop; \
 			     (b).saved_bar = LD->mark_bar; \
-			     DEBUG(CHK_SECURE, assert((b).saved_bar >= gBase && \
-						      (b).saved_bar <= gTop)); \
-			     LD->mark_bar = (b).globaltop = gTop; \
+			     DEBUG(CHK_SECURE, \
+				   assert((b).saved_bar == NO_MARK_BAR || \
+					  ((b).saved_bar >= gBase && \
+					   (b).saved_bar <= gTop))); \
+			     (b).globaltop = gTop; \
+			     if ( LD->mark_bar != NO_MARK_BAR ) \
+			       LD->mark_bar = (b).globaltop; \
 			   } while(0)
 #define DiscardMark(b)	do { LD->mark_bar = (LD->frozen_bar > (b).saved_bar ? \
 					     LD->frozen_bar : (b).saved_bar); \
-			     DEBUG(CHK_SECURE, assert(LD->mark_bar >= gBase && \
-						      LD->mark_bar <= gTop)); \
+			     DEBUG(CHK_SECURE, \
+				   assert(LD->mark_bar == NO_MARK_BAR || \
+					  (LD->mark_bar >= gBase && \
+					   LD->mark_bar <= gTop))); \
 			   } while(0)
 #define NOT_A_MARK	(TrailEntry)(~(word)0)
 #define NoMark(b)	do { (b).trailtop = NOT_A_MARK; \
@@ -1814,7 +1843,6 @@ Temporary store/restore pointers to make them safe over GC/shift
 #define SIG_PROLOG_OFFSET	32	/* Start of Prolog signals */
 
 typedef RETSIGTYPE (*handler_t)(int);
-typedef void *SignalContext;		/* struct sigcontext on sun */
 
 typedef struct
 { handler_t   saved_handler;		/* Original handler */
@@ -1853,14 +1881,6 @@ typedef enum pl_event_type
   PL_EV_THREADFINISHED			/* A thread has finished */
 } pl_event_type;
 
-#ifdef O_DEBUGGER
-#define callEventHook(...) \
-        ( PROCEDURE_event_hook1->definition->impl.any ? \
-		PL_call_event_hook(__VA_ARGS__) : TRUE )
-#else
-#define callEventHook(...) TRUE
-#endif
-
 
 		 /*******************************
 		 *	       COMPARE		*
@@ -2042,9 +2062,10 @@ typedef struct
 		 *	      WAKEUP		*
 		 *******************************/
 
-#define WAKEUP_STATE_WAKEUP    0x1
-#define WAKEUP_STATE_EXCEPTION 0x2
-#define WAKEUP_STATE_SKIP_EXCEPTION 0x4
+#define WAKEUP_STATE_WAKEUP          0x1 /* State contains a wakeup */
+#define WAKEUP_STATE_EXCEPTION	     0x2 /* State contains an exception */
+#define WAKEUP_STATE_SKIP_EXCEPTION  0x4 /* Do not restore exception from state */
+#define WAKEUP_KEEP_URGENT_EXCEPTION 0x8 /* Keep the most urgent exception */
 
 typedef struct wakeup_state
 { fid_t		fid;			/* foreign frame reference */
@@ -2261,6 +2282,8 @@ typedef struct debuginfo
 #define PLFLAG_DOT_IN_ATOM	    0x00800000 /* Allow atoms a.b.c */
 #define PLFLAG_VARPREFIX	    0x01000000 /* Variable must start with _ */
 #define PLFLAG_PROTECT_STATIC_CODE  0x02000000 /* Deny clause/2 on static code */
+#define PLFLAG_ERROR_AMBIGUOUS_STREAM_PAIR 0x04000000
+#define PLFLAG_GCTHREAD		    0x08000000 /* Do atom/clause GC in a thread */
 
 typedef struct
 { unsigned int flags;		/* Fast access to some boolean Prolog flags */
@@ -2361,10 +2384,6 @@ decrease).
 #include "os/pl-string.h"		/* Basic string functions */
 #include "pl-ressymbol.h"		/* Meta atom handling */
 
-#ifdef ATOMIC_INC
-#define ATOMIC_REFERENCES 1		/* Use atomic +/- for atom references */
-#endif
-
 #ifdef __DECC				/* Dec C-compiler: avoid conflicts */
 #undef leave
 #undef except
