@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2014, University of Amsterdam
+    Copyright (c)  1985-2017, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -40,10 +40,12 @@
 #include "os/pl-cstack.h"
 #include "pl-dbref.h"
 #include "pl-trie.h"
+#include "pl-tabling.h"
 #include <sys/stat.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#include <errno.h>
 
 #undef max
 #define max(a,b) ((a) > (b) ? (a) : (b))
@@ -69,6 +71,10 @@ setupProlog(void)
 { GET_LD
   DEBUG(1, Sdprintf("Starting Heap Initialisation\n"));
 
+#ifdef O_LOGICAL_UPDATE
+  next_global_generation();
+#endif
+
   LD->critical = 0;
   LD->signal.pending[0] = 0;
   LD->signal.pending[1] = 0;
@@ -83,10 +89,7 @@ setupProlog(void)
   initForeign();
 #if HAVE_SIGNAL
   DEBUG(1, Sdprintf("Prolog Signal Handling ...\n"));
-  if ( truePrologFlag(PLFLAG_SIGNALS) )
-  { initSignals();
-    initBackTrace();
-  }
+  initSignals();
 #endif
   DEBUG(1, Sdprintf("Stacks ...\n"));
   if ( !initPrologStacks(GD->options.localSize,
@@ -158,6 +161,10 @@ initPrologLocalData(ARG1_LD)
   LD->break_level = -1;
   LD->prolog_flag.write_attributes = PL_WRT_ATTVAR_IGNORE;
 
+#ifdef O_PLMT
+  simpleMutexInit(&LD->thread.scan_lock);
+#endif
+
   updateAlerted(LD);
 }
 
@@ -166,17 +173,25 @@ initPrologLocalData(ARG1_LD)
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 			   SIGNAL HANDLING
 
-SWI-Prolog catches a number of signals.   Interrupt  is catched to allow
-the user to interrupt  normal   execution.  Segmentation  violations are
-trapped on machines using the MMU to implement stack overflow checks and
-stack expansion. These signal handlers needs  to be preserved over saved
-states and the system  should  allow   foreign  language  code to handle
-signals without interfering  with  Prologs   signal  handlers.  For this
-reason a layer is wired around the OS signal handling.
+SWI-Prolog catches a number of signals:
+
+  - SIGINT is caught to allow the user to interrupt normal execution.
+  - SIGUSR2 is caught using an empty handler to break blocking system
+    calls and allow handling of Prolog signals from them.
+  - SIGTERM, SIGABRT and SIGQUIT are caught to cleanup before killing
+    the process again using the same signal.
+  - SIGSEGV, SIGILL, SIGBUS, SIGFPE and SIGSYS are caught by
+    os/pl-cstack.c to print a backtrace and exit.
+  - SIGHUP is caught and causes the process to exit with status 2 after
+    cleanup.
+
+If the system is started using --nosignals, only SIGUSR2 is modified.
+
+Note that library(time) uses SIGUSR1.
 
 Code in SWI-Prolog should  call  PL_signal()  rather  than  signal()  to
 install  signal  handlers.  SWI-Prolog assumes the handler function is a
-void function.  On some systems this gives  some  compiler  warnigns  as
+void function.  On some systems this gives  some  compiler  warnings  as
 they  define  signal handlers to be int functions.  This should be fixed
 some day.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
@@ -184,10 +199,7 @@ some day.
 #if HAVE_SIGNAL
 #define HAVE_SIGNALS 1
 
-#define PLSIG_PREPARED 0x01		/* signal is prepared */
-#define PLSIG_THROW    0x02		/* throw signal(num, name) */
-#define PLSIG_SYNC     0x04		/* call synchronously */
-#define PLSIG_NOFRAME  0x08		/* Do not create a Prolog frame */
+#define PLSIG_PREPARED 0x00010000	/* signal is prepared */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Define the signals and  their  properties.   This  could  be  nicer, but
@@ -279,14 +291,21 @@ static struct signame
 #ifdef SIGPWR
   { SIGPWR,	"pwr",    0},
 #endif
+
+/* The signals below here are recorded as Prolog interrupts, but
+   not supported by OS signals.  They start at offset 32.
+*/
+
   { SIG_EXCEPTION,     "prolog:exception",     0 },
 #ifdef SIG_ATOM_GC
-  { SIG_ATOM_GC,   "prolog:atom_gc",       0 },
+  { SIG_ATOM_GC,       "prolog:atom_gc",       0 },
 #endif
   { SIG_GC,	       "prolog:gc",	       0 },
 #ifdef SIG_THREAD_SIGNAL
   { SIG_THREAD_SIGNAL, "prolog:thread_signal", 0 },
 #endif
+  { SIG_CLAUSE_GC,     "prolog:clause_gc",     0 },
+  { SIG_PLABORT,       "prolog:abort",         0 },
 
   { -1,		NULL,     0}
 };
@@ -379,10 +398,10 @@ to the main thread.
 #undef LD
 #define LD LOCAL_LD
 
-static void
+void
 dispatch_signal(int sig, int sync)
 { GET_LD
-  SigHandler sh = &GD->sig_handlers[sig-1];
+  SigHandler sh = &GD->signals.handlers[sig-1];
   fid_t fid;
   term_t lTopSave;
   int saved_current_signal;
@@ -565,7 +584,7 @@ set_sighandler(int sig, handler_t func)
 
 static SigHandler
 prepareSignal(int sig)
-{ SigHandler sh = &GD->sig_handlers[sig-1];
+{ SigHandler sh = &GD->signals.handlers[sig-1];
 
   if ( false(sh, PLSIG_PREPARED) )
   { set(sh, PLSIG_PREPARED);
@@ -579,7 +598,7 @@ prepareSignal(int sig)
 
 static void
 unprepareSignal(int sig)
-{ SigHandler sh = &GD->sig_handlers[sig-1];
+{ SigHandler sh = &GD->signals.handlers[sig-1];
 
   if ( true(sh, PLSIG_PREPARED) )
   { if ( sig < SIG_PROLOG_OFFSET )
@@ -697,48 +716,66 @@ with EINTR and thus make them interruptable for thread-signals.
 #ifdef SIG_ALERT
 static void
 alert_handler(int sig)
-{ (void)sig;
+{ SigHandler sh = &GD->signals.handlers[sig-1];
+
+  if ( sh->saved_handler &&
+       sh->saved_handler != SIG_IGN &&
+       sh->saved_handler != SIG_DFL )
+    (*sh->saved_handler)(sig);
 }
 #endif
 
 
 static void
 initSignals(void)
-{ struct signame *sn = signames;
+{ GET_LD
 
+  /* This is general signal handling that is not strictly needed */
+  if ( truePrologFlag(PLFLAG_SIGNALS) )
+  { struct signame *sn = signames;
 #ifdef SIGPIPE
-  set_sighandler(SIGPIPE, SIG_IGN);
+    set_sighandler(SIGPIPE, SIG_IGN);
 #endif
-  initTerminationSignals();
-  for( ; sn->name; sn++)
-  {
+    initTerminationSignals();
+    initBackTrace();
+    for( ; sn->name; sn++)
+    {
 #ifdef HAVE_BOEHM_GC
-    if ( sn->sig == GC_get_suspend_signal() ||
-	 sn->sig == GC_get_thr_restart_signal() )
-      sn->flags = 0;
+      if ( sn->sig == GC_get_suspend_signal() ||
+	   sn->sig == GC_get_thr_restart_signal() )
+	sn->flags = 0;
 #endif
-    if ( sn->flags )
-    { SigHandler sh = prepareSignal(sn->sig);
-      sh->flags |= sn->flags;
+      if ( sn->flags )
+      { SigHandler sh = prepareSignal(sn->sig);
+	sh->flags |= sn->flags;
+      }
     }
+
+#ifdef SIGHUP
+    PL_signal(SIGHUP|PL_SIGSYNC, hupHandler);
+#endif
   }
 
-  PL_signal(SIG_EXCEPTION|PL_SIGSYNC, sig_exception_handler);
-  PL_signal(SIG_GC|PL_SIGSYNC, gc_handler);
-  PL_signal(SIG_CLAUSE_GC|PL_SIGSYNC, cgc_handler);
-  PL_signal(SIG_PLABORT|PL_SIGSYNC, abort_handler);
+  /* We do need alerting to make thread signals work while the */
+  /* system is blocked in a system call. Can be controlled with --sigalert=N */
 
 #ifdef SIG_ALERT
-  PL_signal(SIG_ALERT|PL_SIGNOFRAME, alert_handler);
+  if ( GD->signals.sig_alert )
+    PL_signal(GD->signals.sig_alert|PL_SIGNOFRAME, alert_handler);
 #endif
+
+  /* these signals are not related to Unix signals and can thus */
+  /* be enabled always */
+
+  PL_signal(SIG_EXCEPTION|PL_SIGSYNC,     sig_exception_handler);
+  PL_signal(SIG_GC|PL_SIGSYNC,	          gc_handler);
+  PL_signal(SIG_CLAUSE_GC|PL_SIGSYNC,     cgc_handler);
+  PL_signal(SIG_PLABORT|PL_SIGSYNC,       abort_handler);
 #ifdef SIG_THREAD_SIGNAL
   PL_signal(SIG_THREAD_SIGNAL|PL_SIGSYNC, executeThreadSignals);
 #endif
 #ifdef SIG_ATOM_GC
-  PL_signal(SIG_ATOM_GC|PL_SIGSYNC, agc_handler);
-#endif
-#ifdef SIGHUP
-  PL_signal(SIGHUP|PL_SIGSYNC, hupHandler);
+  PL_signal(SIG_ATOM_GC|PL_SIGSYNC,       agc_handler);
 #endif
 }
 
@@ -877,47 +914,100 @@ void blockSignal(int sig) {}
 
 #endif
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+BUG: The interface of PL_signal() is broken   as  it does not return the
+current flags associated with the signal and therefore we cannot restore
+the signal safely. We should  design  a   struct  based  API  similar to
+sigaction().
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+int
+PL_sigaction(int sig, pl_sigaction_t *act, pl_sigaction_t *old)
+{ SigHandler sh = NULL;
+
+  if ( sig < 0 || sig > MAXSIGNAL )
+  { errno = EINVAL;
+    return -1;
+  }
+
+  if ( sig == 0 )
+  { for(sig=SIG_PROLOG_OFFSET; sig<MAXSIGNAL; sig++)
+    { sh = &GD->signals.handlers[sig-1];
+      if ( sh->flags == 0 )
+	break;
+    }
+    if ( !sh )
+    { errno = EBUSY;
+      return -2;
+    }
+  } else
+  { sh = &GD->signals.handlers[sig-1];
+  }
+
+  if ( old )
+  { memset(old, 0, sizeof(*old));
+    old->sa_cfunction = sh->handler;
+    old->sa_predicate = sh->predicate;
+    old->sa_flags     = sh->flags;
+  }
+
+  if ( act && act != old )
+  { int active = FALSE;
+
+    if ( (act->sa_flags&PLSIG_THROW) || act->sa_predicate )
+    { if ( ((act->sa_flags&PLSIG_THROW) && act->sa_predicate) ||
+	   act->sa_cfunction )
+      { errno = EINVAL;
+	return -1;
+      }
+      active = TRUE;
+    } else if ( act->sa_cfunction &&
+		(false(sh, PLSIG_PREPARED)||act->sa_cfunction!=sh->saved_handler) )
+    { active = TRUE;
+    }
+
+    if ( active )
+    { sh->handler   = act->sa_cfunction;
+      sh->predicate = act->sa_predicate;
+      sh->flags     = (sh->flags&~0xffff)|act->sa_flags;
+      if ( false(sh, PLSIG_PREPARED) )
+	prepareSignal(sig);
+    } else
+    { unprepareSignal(sig);
+      sh->handler   = NULL;
+      sh->predicate = NULL;
+      sh->flags     = 0;
+    }
+  }
+
+  return sig;
+}
 
 
 handler_t
 PL_signal(int sigandflags, handler_t func)
-{ if ( HAVE_SIGNALS )
-  { handler_t old;
-    SigHandler sh;
-    int sig = (sigandflags & 0xffff);
+{
+#ifdef HAVE_SIGNALS
+  pl_sigaction_t act = {0};
+  pl_sigaction_t old;
 
-    if ( sig > MAXSIGNAL )
-    { warning("PL_signal(): illegal signal number: %d", sig);
-      return SIG_DFL;
-    }
+  act.sa_cfunction = func;
+  if ( (sigandflags&PL_SIGSYNC) )
+    act.sa_flags |= PLSIG_SYNC;
+  if ( (sigandflags&PL_SIGNOFRAME) )
+    act.sa_flags |= PLSIG_NOFRAME;
 
-    sh = &GD->sig_handlers[sig-1];
-    if ( true(sh, PLSIG_PREPARED) )
-    { old = sh->handler;
-      if ( func == sh->saved_handler )
-	unprepareSignal(sig);
-      else
-	sh->handler = func;
-    } else
-    { sh = prepareSignal(sig);
-      old = sh->saved_handler;
-      sh->handler = func;
-    }
-    if ( func != SIG_DFL )
-      clear(sh, PLSIG_THROW);		/* we have a user handler now */
+  if ( PL_sigaction((sigandflags & 0xffff), &act, &old) >= 0 )
+  { if ( (old.sa_flags&PLSIG_PREPARED) && old.sa_cfunction )
+      return old.sa_cfunction;
 
-    if ( (sigandflags & PL_SIGSYNC) )
-      set(sh, PLSIG_SYNC);
-    else
-      clear(sh, PLSIG_SYNC);
-    if ( (sigandflags & PL_SIGNOFRAME) )
-      set(sh, PLSIG_NOFRAME);
-    else
-      clear(sh, PLSIG_NOFRAME);
-
-    return old;
-  } else
     return SIG_DFL;
+  }
+
+  return NULL;
+#else
+  return SIG_DFL;
+#endif
 }
 
 
@@ -1043,7 +1133,7 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
   } else
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_signal, sig);
 
-  sh = &GD->sig_handlers[sign-1];
+  sh = &GD->signals.handlers[sign-1];
 
   if ( false(sh, PLSIG_PREPARED) )		/* not handled */
   { TRY(PL_unify_atom(old, ATOM_default));
@@ -1280,11 +1370,18 @@ allocStacks(size_t local, size_t global, size_t trail)
   global   = max(global,   minglobal);
   trail    = max(trail,    mintrail);
 
+  gBase = NULL;
+  tBase = NULL;
+  aBase = NULL;
+
   gBase = (Word)       stack_malloc(iglobal + ilocal);
   tBase = (TrailEntry) stack_malloc(itrail);
   aBase = (Word *)     stack_malloc(minargument);
+
   if ( !gBase || !tBase || !aBase )
-  { freeStacks(PASS_LD1);
+  { if ( gBase )
+      *gBase++ = MARK_MASK;		/* compensate for freeStacks */
+    freeStacks(PASS_LD1);
     fail;
   }
 
@@ -1336,10 +1433,8 @@ stack_malloc(size_t size)
 #ifdef SECURE_GC
     memset(sp, 0xFB, size);
 #endif
+    ATOMIC_ADD(&GD->statistics.stack_space, size);
 
-    PL_LOCK(L_MISC);
-    GD->statistics.stack_space += size;
-    PL_UNLOCK(L_MISC);
     return sp;
   }
 
@@ -1362,10 +1457,10 @@ stack_realloc(void *old, size_t size)
   if ( (mem = realloc(sp, size+sizeof(size_t))) )
   { sp = mem;
     *sp++ = size;
-    PL_LOCK(L_MISC);
-    GD->statistics.stack_space -= osize;
-    GD->statistics.stack_space += size;
-    PL_UNLOCK(L_MISC);
+    if ( size > osize )
+      ATOMIC_ADD(&GD->statistics.stack_space, size-osize);
+    else
+      ATOMIC_SUB(&GD->statistics.stack_space, osize-size);
     return sp;
   }
 #endif
@@ -1378,9 +1473,7 @@ stack_free(void *mem)
 { size_t *sp = mem;
   size_t osize = *--sp;
 
-  PL_LOCK(L_MISC);
-  GD->statistics.stack_space -= osize;
-  PL_UNLOCK(L_MISC);
+  ATOMIC_SUB(&GD->statistics.stack_space, osize);
 
 #ifdef SECURE_GC
   memset(sp, 0xFB, osize+sizeof(size_t));
@@ -1561,6 +1654,8 @@ freePrologLocalData(PL_local_data_t *ld)
 
   if ( ld->qlf.getstr_buffer )
     free(ld->qlf.getstr_buffer);
+
+  clearThreadTablingData(ld);
 }
 
 
